package com.bureau.devicefingerprint

import android.Manifest
import android.app.Application
import android.content.Context
import android.content.pm.PackageManager
import android.util.Log
import androidx.annotation.Keep
import androidx.core.app.ActivityCompat
import com.bureau.base.AppInfo
import com.bureau.base.DevUtils
import com.bureau.base.Environment
import com.bureau.base.LocalErrorCodes
import com.bureau.behavioralbiometrics.UserBehaviorCapture
import com.bureau.devicefingerprint.data.remote.RemoteDataSource
import com.bureau.devicefingerprint.datacollectors.SystemPropertiesProvider
import com.bureau.devicefingerprint.di.DaggerDeviceFingerprintComponent
import com.bureau.devicefingerprint.di.DeviceFingerprintComponent
import com.bureau.devicefingerprint.models.Callback
import com.bureau.devicefingerprint.models.CredentialBuilder
import com.bureau.devicefingerprint.models.ErrorResponse
import com.bureau.devicefingerprint.models.PayLoadCallback
import com.bureau.devicefingerprint.models.SubmitResponse
import com.bureau.devicefingerprint.models.SuspiciousSignalCallback
import com.bureau.devicefingerprint.models.devicedataholder.CodecInfo
import com.bureau.devicefingerprint.models.devicedataholder.DeviceDataHolder
import com.bureau.devicefingerprint.models.devicedataholder.MediaDrmId
import com.bureau.devicefingerprint.models.devicedataholder.RealTimeDatHolder
import com.bureau.devicefingerprint.tools.DataCallback
import com.bureau.devicefingerprint.tools.RealTimeDataCallback
import com.bureau.devicefingerprint.tools.RealTimeDataChannel
import com.bureau.devicefingerprint.tools.SharedPreferenceHelper
import com.bureau.devicefingerprint.tools.TimeLoggerUtil
import com.bureau.devicefingerprint.usecases.GetEncryptedDeviceDataUseCase
import io.sentry.Sentry
import io.sentry.SentryOptions
import kotlinx.coroutines.CoroutineScope
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.launch
import javax.inject.Inject

@Keep
class BureauAPI private constructor(private val application: Application, private val clientId: String, private val sessionId: String, environment: Environment, enableBehavioralBiometrics: Boolean, isDev: Boolean = false) {
    @Inject
    internal lateinit var remoteDataSource: RemoteDataSource
    private var isStartSessionCalled = false
    private var credentialBuilder: CredentialBuilder? = CredentialBuilder()
    private val deviceData: DeviceDataHolder = DeviceDataHolder()

    @Inject
    internal lateinit var dataMapper: DataMapper

    @Inject
    internal lateinit var appInfo: AppInfo

    @Inject
    internal lateinit var dataMonitoring: DataMonitoring
    private var permissionList: ArrayList<String> = arrayListOf(Manifest.permission.ACCESS_FINE_LOCATION, Manifest.permission.ACCESS_COARSE_LOCATION)
    private var isRealTimeDataEnabled: Boolean = false
    private var realTimeDataHolder: RealTimeDatHolder? = null
    private var realTimeDataCallback: RealTimeDataCallback? = null
    private var userBehaviorCapture: UserBehaviorCapture? = null

    @Inject
    internal lateinit var getEncryptedDeviceDataUseCase: GetEncryptedDeviceDataUseCase
    private var deviceFingerprintComponent: DeviceFingerprintComponent

    @Inject
    internal lateinit var systemPropertiesProvider: SystemPropertiesProvider

    init {
        TimeLoggerUtil.timeLoggerUtil.logTrace("SDKLoaded")
        deviceFingerprintComponent = DaggerDeviceFingerprintComponent.factory().create(application, environment, clientId, sessionId, isDev, enableBehavioralBiometrics).apply {
            inject(this@BureauAPI)
        }
        if (!Sentry.isEnabled()) {
            sendExceptionToSentry = true
            Sentry.init { options: SentryOptions ->
                options.dsn = BuildConfig.SENTRY_DSN_DEVICE
                // Set tracesSampleRate to 1.0 to capture 100% of transactions for performance monitoring.
                // We recommend adjusting this value in production.
                options.tracesSampleRate = 1.0
                // When first trying Sentry it's good to see what the SDK is doing:
                // options.isDebug = true
                options.environment = if (BuildConfig.DEBUG) "SANDBOX" else "PRODUCTION"
                options.setTag("isDebug", BuildConfig.DEBUG.toString())
                options.setTag("buildType", BuildConfig.BUILD_TYPE)
                options.setTag("libraryPackageName", BuildConfig.LIBRARY_PACKAGE_NAME)
                options.setTag("libraryVersion", BuildConfig.LIBRARY_VERSION_NAME)
                options.release = BuildConfig.LIBRARY_VERSION_NAME
                options.proguardUuid = "d14834aa-8f6e-11ee-b9d1-0242ac120002"
                options.addBundleId("f2e831c4-8f59-11ee-b9d1-0242ac120002")
            }
            setupGlobalTags()
        }
        if (clientId.isEmpty()) {
            throw Error("Credential id should not be empty")
        }

        credentialBuilder?.sessionId = sessionId

        if (enableBehavioralBiometrics) {
            userBehaviorCapture = UserBehaviorCapture.init(application, environment, clientId, sessionId, sendExceptionToSentry, isDev)
        }
        TimeLoggerUtil.timeLoggerUtil.logTrace("SDKInitDone")
    }

    private fun setupGlobalTags() {
        Sentry.setTag("client.appPackage", appInfo.getPackageName())
        Sentry.setTag("client.appVersion", appInfo.getVersion())
        Sentry.setTag("client.appName", appInfo.getAppName())
        Sentry.setTag("appBuildType", if (appInfo.isInDebugMode()) "debug" else "release")
    }

    fun startSubSession(subSessionId: String) {
        try {
            setUpMandatoryDataCollection()
            isStartSessionCalled = true
            userBehaviorCapture?.startSubSession(subSessionId, credentialBuilder?.userId ?: "")
        } catch (throwable: Throwable) {
            handleException(throwable)
        }
    }

    fun setUserId(userId: String) {
        try {
            credentialBuilder?.userId = userId
        } catch (throwable: Throwable) {
            handleException(throwable)
        }
    }

    fun setFlow(flow: String) {
        try {
            credentialBuilder?.flow = flow
        } catch (throwable: Throwable) {
            handleException(throwable)
        }
    }

    fun stopSubSession() {
        try {
            userBehaviorCapture?.stopSubSession()
        } catch (throwable: Throwable) {
            handleException(throwable)
        }
    }

    private fun collectData() {
        var hasLocationPermission = false
        if (hasPermissions(application, permissionList)) {
            hasLocationPermission = true
        }
        dataMapper.getSynchronisedOSData(hasLocationPermission, deviceData)
    }

    fun setRealtimeCallback(callback: RealTimeDataCallback) {
        try {
            this.realTimeDataCallback = callback
            isRealTimeDataEnabled = true
            realTimeDataHolder = RealTimeDatHolder()
            Log.i("DeviceService", "Real Time Data Monitoring Started")
            dataMonitoring.registerDataMonitoring(object : RealTimeDataChannel {
                override fun emitVPNStatus(item: Boolean) {
                    realTimeDataHolder?.vpnStatus_ = item
                    if (item) {
                        SharedPreferenceHelper.setSharedPreferenceBoolean(
                            application,
                            SharedPreferenceHelper.isPastVPNData,
                            value = true,
                            isDataMonitoring = true
                        )
                    }
                    CoroutineScope(Dispatchers.Main).launch {
                        realTimeDataCallback?.onVPNStatusChange(item)
                    }
                }

                override fun emitMockGPSStatus(item: Boolean) {
                    realTimeDataHolder?.mockGPSStatus_ = item
                    if (item) {
                        SharedPreferenceHelper.setSharedPreferenceBoolean(
                            application,
                            SharedPreferenceHelper.isPastMockGPSData,
                            value = true,
                            isDataMonitoring = true
                        )
                    }
                    CoroutineScope(Dispatchers.Main).launch {
                        realTimeDataCallback?.onMockGpsStatusChange(item)
                    }
                }

                override fun emitDebuggerStatus(item: Boolean) {
                    realTimeDataHolder?.debuggerStatus_ = item
                    if (item) {
                        SharedPreferenceHelper.setSharedPreferenceBoolean(
                            application,
                            SharedPreferenceHelper.isPastDebuggerData,
                            value = true,
                            isDataMonitoring = true
                        )
                    }
                    CoroutineScope(Dispatchers.Main).launch {
                        realTimeDataCallback?.onDebuggerStatusChange(item)
                    }
                }
            })
        } catch (throwable: Throwable) {
            handleException(throwable)
        }
    }

    fun submit(callback: DataCallback, suspiciousSignalCallback: SuspiciousSignalCallback? = null) {
        TimeLoggerUtil.timeLoggerUtil.logTrace("CalledSubmit")
        try {
            if (sessionId.isEmpty()) {
                callback.onError(
                    ErrorResponse(
                        "Error : SessionId should not be empty",
                        LocalErrorCodes.ERROR_CODE_SESSION_ID_EMPTY.toString()
                    )
                )
            } else {
                stopSubSession()
                submitData(callback, null, suspiciousSignalCallback)
            }
        } catch (throwable: Throwable) {
            handleException(throwable)
            callback.onError(
                ErrorResponse(
                    "Fatal error",
                    LocalErrorCodes.ERROR_CODE_GENERAL_THROWABLE.toString()
                )
            )
        }
    }

    fun getPayLoadForSubmit(payLoadCallback: PayLoadCallback, suspiciousSignalCallback: SuspiciousSignalCallback? = null) {
        try {
            if (sessionId.isEmpty()) {
                payLoadCallback.onError(
                    ErrorResponse(
                        "Error : SessionId should not be empty",
                        LocalErrorCodes.ERROR_CODE_SESSION_ID_EMPTY.toString()
                    )
                )
            } else {
                stopSubSession()
                submitData(null, payLoadCallback, suspiciousSignalCallback)
            }
        } catch (throwable: Throwable) {
            handleException(throwable)
            payLoadCallback.onError(
                ErrorResponse(
                    "Fatal error",
                    LocalErrorCodes.ERROR_CODE_GENERAL_THROWABLE.toString()
                )
            )
        }
    }

    private fun submitData(dataCallback: DataCallback?, payLoadCallback: PayLoadCallback?, suspiciousSignalCallback: SuspiciousSignalCallback?) {
        if (!isStartSessionCalled) {
            setUpMandatoryDataCollection()
        }

        CoroutineScope(Dispatchers.IO).launch {
            val job1 = CoroutineScope(Dispatchers.IO).launch {
                deviceData.let {
                    dataMapper.getAsyncOSData(
                        credentialBuilder = credentialBuilder,
                        it
                    )
                }
            }

            var mediaDrmId: MediaDrmId? = null
            val job2 = CoroutineScope(Dispatchers.IO).launch {
                mediaDrmId = dataMapper.getMediaDrmId()
            }

            var codecInfo: CodecInfo? = null
            val job3 = CoroutineScope(Dispatchers.IO).launch {
                codecInfo = dataMapper.getMediaCodecInfo()
            }

            var isHookingDetected = false
            val job4 = CoroutineScope(Dispatchers.IO).launch {
                isHookingDetected = dataMapper.getIsHookingDetected()
            }

            job1.join()
            job2.join()
            job3.join()
            job4.join()
            deviceData.mediaDrmId_ = mediaDrmId
            deviceData.codecInfo_ = codecInfo
            deviceData.isHookingDetected = isHookingDetected

            deviceData.let { dataMapper.getAsynchronousIds(it) }
            deviceData.let { dataMapper.addGeocodeData(it) }
            if (isRealTimeDataEnabled) {
                deviceData.realTimeDataInfo_ = realTimeDataHolder
            }

            sendData(dataCallback, payLoadCallback, suspiciousSignalCallback)
        }
    }

    private suspend fun sendData(dataCallback: DataCallback?, payLoadCallback: PayLoadCallback?, suspiciousSignalCallback: SuspiciousSignalCallback?) {
        deviceData.let { dataMapper.addPastRealData(it) }

        if (sessionId.isNotEmpty()) {
            deviceData.sessionId_ = sessionId
        }

        TimeLoggerUtil.timeLoggerUtil.logTrace("CallingEncryptionAndSendData")
        deviceData.metaData.SDKEvents = TimeLoggerUtil.timeLoggerUtil.getEventMap()
        TimeLoggerUtil.timeLoggerUtil.endTrace()

        handleSuspiciousSignals(suspiciousSignalCallback)

        val encryptedDeviceData = getEncryptedDeviceDataUseCase(deviceData)
        if (encryptedDeviceData.eaIV.isEmpty() || encryptedDeviceData.eaKey.isEmpty()) {
            dataCallback?.onError(ErrorResponse("invalid data", LocalErrorCodes.ERROR_CODE_EA_KEYS_EMPTY.toString()))
            return
        }
        if (payLoadCallback != null) {
            payLoadCallback.onResult(encryptedDeviceData, remoteDataSource.getSubmitHeaders(clientId))
            return
        }

        remoteDataSource.sendData(
            encryptedDeviceData = encryptedDeviceData,
            object : Callback {
                override fun getResult(isSuccess: Boolean?) {
                    SharedPreferenceHelper.clearAllDataPref(application)

                    CoroutineScope(Dispatchers.Main).launch {
                        isStartSessionCalled = false
                        dataCallback?.onResult(SubmitResponse("Success"))
                    }
                }

                override fun getError(error: String?, errorCode: String) {
                    SharedPreferenceHelper.clearAllDataPref(application)
                    CoroutineScope(Dispatchers.Main).launch {
                        isStartSessionCalled = false

                        dataCallback?.onError(ErrorResponse(error ?: "Error", errorCode))
                    }
                }
            },
            clientId
        )
    }

    private fun handleSuspiciousSignals(suspiciousSignalCallback: SuspiciousSignalCallback?) {
        suspiciousSignalCallback?.let { signalCallback ->
            if (deviceData.settingsInfo_?.developmentSettingsEnabled_ == "1") {
                signalCallback.onDeveloperModeEnabled()
            }
            if (deviceData.settingsInfo_?.adbEnabled_ == "1") {
                signalCallback.onADBEnabled()
            }
            if (deviceData.debuggerInfo_?.isRooted_ == true) {
                signalCallback.onDeviceRooted()
            }
            if (!systemPropertiesProvider.getIsBootloaderLocked()) {
                signalCallback.onBootLoaderUnlocked()
            }
            if (deviceData.locationInfo_?.isMockGPS_ == true) {
                signalCallback.onMockGpsEnabled()
            }
            if (deviceData.packageManagerInfo_?.isPackageDebuggable_ == true) {
                signalCallback.onPackageDebuggable()
            }

            if (!deviceData.appDirectorypath_.isNullOrEmpty()) {
                val appDirectoryPath = deviceData.appDirectorypath_ ?: ""
                val cloneKeywords = arrayOf("clone", "parallel", "dual", "multi", "virtual")
                for (cloneKeyword in cloneKeywords) {
                    if (appDirectoryPath.contains(cloneKeyword)) {
                        signalCallback.onAppCloningDetected()
                        return@let
                    }
                }

                val allowedAppDirPath = mapOf("/data/user/0/${deviceData.ndkInfo_?.applicationId_}/files" to true, "/data/data/${deviceData.ndkInfo_?.applicationId_}/files" to true)
                if (!allowedAppDirPath.containsKey(appDirectoryPath)) {
                    signalCallback.onAppCloningDetected()
                    return@let
                }
            }
        }
    }

    private fun checkErrors() {
        if (sessionId.isEmpty()) {
            throw Error("Session id should not be empty")
        }
        if (credentialBuilder?.userId?.isEmpty() == true) {
            Log.w("Prism SDK", "User Id received is empty")
        }
    }

    private fun setUpMandatoryDataCollection() {
        checkErrors()
        collectData()
    }

    fun removeRealTimeCallBack() {
        try {
            if (isRealTimeDataEnabled) {
                this.realTimeDataCallback = null
                isRealTimeDataEnabled = false
                dataMonitoring.unRegisterDataMonitoring()
                Log.i("DeviceService", "Real Time Data Monitoring Stopped")
            }
        } catch (throwable: Throwable) {
            handleException(throwable)
        }
    }

    @Keep
    companion object {

        private var sendExceptionToSentry = false
        fun init(application: Application, clientId: String, sessionId: String, environment: Environment, enableBehavioralBiometrics: Boolean = false): BureauAPI {
            TimeLoggerUtil.timeLoggerUtil.startTrace()
            return BureauAPI(application, clientId, sessionId, environment, enableBehavioralBiometrics, DevUtils.isDev)
        }

        fun hasPermissions(context: Context, permissionList: List<String>): Boolean {
            permissionList.forEach {
                if (ActivityCompat.checkSelfPermission(
                        context,
                        it
                    ) != PackageManager.PERMISSION_GRANTED
                ) {
                    return false
                }
            }
            return true
        }

        fun hasReadPhoneStatePermission(context: Context): Boolean {
            return hasPermissions(context, listOf(Manifest.permission.READ_PHONE_STATE))
        }

        internal fun handleException(throwable: Throwable) {
            try {
                if (sendExceptionToSentry) {
                    Sentry.captureException(throwable)
                }
            } catch (throwable: Throwable) {
                Log.e("Sentry", "Failed to report a crash.", throwable)
            }
            /*if (BuildConfig.DEBUG) {
                Log.e("Sentry", "Something went wrong.", throwable)
                throw RuntimeException(throwable)
            }*/
        }
    }
}
