package com.bureau.devicefingerprint.data.remote

import android.util.Log
import com.bureau.base.Environment
import com.bureau.devicefingerprint.BureauAPI
import com.bureau.devicefingerprint.NetworkService
import com.bureau.devicefingerprint.di.DeviceFingerprintScope
import com.bureau.devicefingerprint.di.IS_DEV
import com.bureau.devicefingerprint.models.Callback
import com.bureau.devicefingerprint.models.EncryptedDeviceData
import com.bureau.devicefingerprint.tools.ENCRYPTION_VERSION
import com.bureau.devicefingerprint.tools.SystemPropertiesInvoker
import com.bureau.network.RetroFitInstance.getRetrofitInstance
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.withContext
import retrofit2.HttpException
import javax.inject.Inject
import javax.inject.Named

@DeviceFingerprintScope
class RemoteDataSourceImpl @Inject constructor(environment: Environment, @Named(IS_DEV) isDev: Boolean) : RemoteDataSource {
    private val baseUrl = if (isDev) {
        BASE_URl_DEV
    } else {
        if (environment == Environment.ENV_PRODUCTION) {
            BASE_URl_PROD
        } else {
            BASE_URl_STG
        }
    }
    private val service: NetworkService by lazy {
        getRetrofitInstance(baseUrl).create(
            NetworkService::class.java
        )
    }

    override suspend fun sendData(encryptedDeviceData: EncryptedDeviceData, callback: Callback?, merchantId: String) = withContext(Dispatchers.IO) {
        try {
            val response = service.sendDeviceDataAsync(encryptedDeviceData, getSubmitHeaders(merchantId))

            if (response.isSuccessful) {
                callback?.getResult(true)
            } else {
                val error: String = try {
                    val errorMap =
                        SystemPropertiesInvoker.jsonToMap(response.errorBody()?.string() ?: "")
                    SystemPropertiesInvoker.jsonToMap(errorMap["error"])["description"] ?: ""
                } catch (e: java.lang.Exception) {
                    BureauAPI.handleException(e)
                    ""
                }
                if (error.isEmpty()) {
                    Log.w("DeviceServiceError", " Api Error" + response.errorBody()?.string())
                    Log.w("DeviceServiceError", "Error Code " + response.code())
                }
                callback?.getError("Error: $error", response.code().toString())
            }
        } catch (ex: HttpException) {
            BureauAPI.handleException(ex)
            callback?.getError("Network Error: Http Exception", "420")
        } catch (ex: Throwable) {
            BureauAPI.handleException(ex)
            callback?.getError("Network Error", "421")
        }
    }

    override suspend fun getSubmitHeaders(merchantId: String): Map<String, String> {
        return mapOf(
            "Content-Type" to "application/json",
            "X-Bureau-Auth-Credential-ID" to merchantId,
            "x-bureau-api-version" to ENCRYPTION_VERSION
        )
    }

    companion object {
        private const val BASE_URl_STG = "https://api.overwatch.stg.bureau.id/"
        private const val BASE_URl_PROD = "https://api.overwatch.bureau.id/"
        private const val BASE_URl_DEV = "https://api.overwatch.dev.bureau.id/"
    }
}
