package com.bureau.devicefingerprint.datacollectors

import android.Manifest
import android.content.Context
import android.net.ConnectivityManager
import android.net.Network
import android.net.NetworkCapabilities
import android.os.Build
import android.telephony.TelephonyManager
import com.bureau.devicefingerprint.BureauAPI
import com.bureau.devicefingerprint.di.DeviceFingerprintScope
import com.bureau.devicefingerprint.tools.SystemPropertiesInvoker
import com.bureau.devicefingerprint.tools.executeSafe
import java.net.Inet4Address
import java.net.Inet6Address
import java.net.InetAddress
import java.net.NetworkInterface
import java.util.*
import javax.inject.Inject

@DeviceFingerprintScope
class NetworkInfoProvider @Inject constructor(var context: Context) {
    fun getIpv6Address(): String? {
        return executeSafe({
            val en = NetworkInterface.getNetworkInterfaces()
            while (en.hasMoreElements()) {
                val intf = en.nextElement()
                val enumIpAddr: Enumeration<InetAddress> = intf.inetAddresses
                while (enumIpAddr.hasMoreElements()) {
                    val inetAddress: InetAddress = enumIpAddr.nextElement()
                    if (!inetAddress.isLoopbackAddress && inetAddress is Inet6Address) {
                        return@executeSafe inetAddress.getHostAddress()
                    }
                }
            }

            ""
        }, "")
    }

    fun getIpv4Address(): String? {
        return executeSafe({
            val en = NetworkInterface.getNetworkInterfaces()
            while (en.hasMoreElements()) {
                val intf = en.nextElement()
                val enumIpAddr: Enumeration<InetAddress> = intf.inetAddresses
                while (enumIpAddr.hasMoreElements()) {
                    val inetAddress: InetAddress = enumIpAddr.nextElement()
                    if (!inetAddress.isLoopbackAddress && inetAddress is Inet4Address) {
                        return@executeSafe inetAddress.getHostAddress()
                    }
                }
            }

            ""
        }, "")
    }

    fun isConnectedToVPN(): Boolean {
        val connectivityManager = context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
        return executeSafe({
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                val activeNetwork: Network? = connectivityManager.activeNetwork
                val caps: NetworkCapabilities? =
                    connectivityManager.getNetworkCapabilities(activeNetwork)
                return@executeSafe caps?.hasTransport(NetworkCapabilities.TRANSPORT_VPN) == true
            } else {
                val networks: Array<Network> = connectivityManager.allNetworks

                for (i in networks.indices) {
                    val caps: NetworkCapabilities? = connectivityManager.getNetworkCapabilities(networks[i])
                    return@executeSafe caps?.hasTransport(NetworkCapabilities.TRANSPORT_VPN) == true
                }
                return@executeSafe false
            }
        }, false)
    }

    fun getWifiMACAddress(): String {
        return executeSafe({
            SystemPropertiesInvoker.getMACAddress("wlan0") ?: ""
        }, "")
    }

    fun getEthernetMACAddress(): String {
        return executeSafe({
            SystemPropertiesInvoker.getMACAddress("eth0") ?: ""
        }, "")
    }

    fun getIsWifiConnected(): Boolean {
        val connectivityManager = context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
        return executeSafe({
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                val networkCapabilities = connectivityManager.getNetworkCapabilities(connectivityManager.activeNetwork)
                return@executeSafe networkCapabilities?.hasTransport(NetworkCapabilities.TRANSPORT_WIFI) == true
            } else {
                val activeNetworkInfo = connectivityManager.activeNetworkInfo
                return@executeSafe activeNetworkInfo != null && activeNetworkInfo.type == ConnectivityManager.TYPE_WIFI
            }
        }, false)
    }

    fun getIsCellularConnected(): Boolean {
        val connectivityManager = context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
        return executeSafe({
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                val networkCapabilities = connectivityManager.getNetworkCapabilities(connectivityManager.activeNetwork)
                return@executeSafe networkCapabilities?.hasTransport(NetworkCapabilities.TRANSPORT_CELLULAR) == true
            } else {
                val activeNetworkInfo = connectivityManager.activeNetworkInfo
                return@executeSafe activeNetworkInfo != null && activeNetworkInfo.type == ConnectivityManager.TYPE_MOBILE
            }
        }, false)
    }

    fun getNetworkType(): String {
        return executeSafe({
            if (BureauAPI.hasPermissions(context, listOf(Manifest.permission.ACCESS_NETWORK_STATE))) {
                val manager =
                    context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
                val networkInfo = manager.getNetworkInfo(ConnectivityManager.TYPE_WIFI)
                if (networkInfo != null && networkInfo.isConnectedOrConnecting) {
                    return@executeSafe NetworkType.WIFI.name
                }
            }

            if (BureauAPI.hasReadPhoneStatePermission(context)) {
                val telephonyManager =
                    context.getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager
                return@executeSafe when (telephonyManager.networkType) {
                    TelephonyManager.NETWORK_TYPE_UNKNOWN -> NetworkType.UNKNOWN.name

                    TelephonyManager.NETWORK_TYPE_GPRS,
                    TelephonyManager.NETWORK_TYPE_EDGE,
                    TelephonyManager.NETWORK_TYPE_CDMA,
                    TelephonyManager.NETWORK_TYPE_1xRTT,
                    TelephonyManager.NETWORK_TYPE_IDEN,
                    TelephonyManager.NETWORK_TYPE_GSM
                    -> NetworkType.CELLULAR_2G.name

                    TelephonyManager.NETWORK_TYPE_UMTS,
                    TelephonyManager.NETWORK_TYPE_EVDO_0,
                    TelephonyManager.NETWORK_TYPE_EVDO_A,
                    TelephonyManager.NETWORK_TYPE_HSDPA,
                    TelephonyManager.NETWORK_TYPE_HSUPA,
                    TelephonyManager.NETWORK_TYPE_HSPA,
                    TelephonyManager.NETWORK_TYPE_EVDO_B,
                    TelephonyManager.NETWORK_TYPE_EHRPD,
                    TelephonyManager.NETWORK_TYPE_HSPAP,
                    TelephonyManager.NETWORK_TYPE_TD_SCDMA
                    -> NetworkType.CELLULAR_3G.name

                    TelephonyManager.NETWORK_TYPE_LTE,
                    TelephonyManager.NETWORK_TYPE_IWLAN
                    -> NetworkType.CELLULAR_4G.name

                    TelephonyManager.NETWORK_TYPE_NR -> NetworkType.CELLULAR_5G.name

                    else -> NetworkType.UNKNOWN.name
                }
            }

            ""
        }, "")
    }

    enum class NetworkType {
        CELLULAR_2G,
        CELLULAR_3G,
        CELLULAR_4G,
        CELLULAR_5G,
        WIFI,
        UNKNOWN
    }
}
