package com.bureau.devicefingerprint.datacollectors

import android.content.Context
import android.content.pm.PackageManager
import com.bureau.devicefingerprint.tools.executeSafe

interface PackageManagerDataSource {
    fun getApplicationsList(): List<PackageInfo>
    fun getSystemApplicationsList(): List<PackageInfo>
    fun getApplicationListInString(): String
    fun getPackageVersion(): String
    fun getClientAppInstallTime(): Long
}

data class PackageInfo(
    val packageName: String,
    val installTime: Long
)

internal class PackageManagerDataSourceImpl(
    private val packageManager: PackageManager,
    private val context: Context
) : PackageManagerDataSource {
    override fun getApplicationsList() = executeSafe(
        {
            packageManager
                .getInstalledPackages(PackageManager.GET_META_DATA)
                .map {
                    PackageInfo(
                        it.packageName,
                        it.firstInstallTime
                    )
                }
        },
        emptyList()
    )

    override fun getSystemApplicationsList() = executeSafe(
        {
            packageManager
                .getInstalledPackages(PackageManager.GET_META_DATA)
                .filter {
                    it.applicationInfo.sourceDir.contains("/system/")
                }
                .map {
                    PackageInfo(
                        it.packageName,
                        it.firstInstallTime
                    )
                }
        },
        emptyList()
    )

    override fun getApplicationListInString(): String {
        val sb = StringBuilder()

        getApplicationsList()
            .sortedBy { it.packageName }
            .forEach { sb.append(it.packageName) }

        return sb.toString()
    }

    override fun getPackageVersion(): String {
        val packageInfo: android.content.pm.PackageInfo? =
            packageManager.getPackageInfo(context.packageName, 0)
        return executeSafe({
            packageInfo?.versionName ?: ""
        }, "")
    }

    override fun getClientAppInstallTime(): Long {
        return executeSafe({
            val packageInfo: android.content.pm.PackageInfo? =
                packageManager.getPackageInfo(context.packageName, 0)
            packageInfo?.firstInstallTime ?: -1
        }, -1)
    }
}
