package com.bureau.devicefingerprint.datacollectors

import android.Manifest
import android.content.Context
import android.os.Build
import android.telephony.SubscriptionManager
import android.telephony.TelephonyManager
import androidx.appcompat.app.AppCompatActivity
import com.bureau.devicefingerprint.BureauAPI
import com.bureau.devicefingerprint.di.DeviceFingerprintScope
import com.bureau.devicefingerprint.tools.executeSafe
import javax.inject.Inject

interface SimInfoProvider {
    fun getNumOfSimSlots(): Int
    fun getNumOfSimsPresent(): Int
    fun isDataRoamingEnabled(): Int
}

@DeviceFingerprintScope
class SimInfoProviderImpl @Inject constructor(private val context: Context) : SimInfoProvider {

    private val telephonyManager = context.getSystemService(AppCompatActivity.TELEPHONY_SERVICE) as TelephonyManager
    private val subscriptionManager = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP_MR1) context.getSystemService(AppCompatActivity.TELEPHONY_SUBSCRIPTION_SERVICE) as SubscriptionManager? else null
    override fun getNumOfSimSlots(): Int {
        return executeSafe({
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.R) {
                telephonyManager.activeModemCount
            } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
                telephonyManager.phoneCount
            } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP_MR1) {
                subscriptionManager?.activeSubscriptionInfoCountMax ?: -1
            } else {
                -1
            }
        }, -1)
    }

    override fun getNumOfSimsPresent(): Int {
        return executeSafe({
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                val simStateMain = telephonyManager.getSimState(0)
                val simStateSecond = telephonyManager.getSimState(1)

                if (TelephonyManager.SIM_STATE_ABSENT != simStateMain && TelephonyManager.SIM_STATE_ABSENT != simStateSecond) {
                    2
                } else if (TelephonyManager.SIM_STATE_ABSENT != simStateMain || TelephonyManager.SIM_STATE_ABSENT != simStateSecond) {
                    1
                } else {
                    0
                }
            } else if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP_MR1) {
                if (BureauAPI.hasReadPhoneStatePermission(context)) {
                    subscriptionManager?.activeSubscriptionInfoCount ?: -1
                } else if (TelephonyManager.SIM_STATE_ABSENT != telephonyManager.simState) {
                    1
                } else {
                    0
                }
            } else if (TelephonyManager.SIM_STATE_ABSENT != telephonyManager.simState) {
                1
            } else {
                0
            }
        }, -1)
    }

    override fun isDataRoamingEnabled(): Int {
        return executeSafe({
            if (BureauAPI.hasReadPhoneStatePermission(context) || BureauAPI.hasPermissions(
                    context,
                    listOf(Manifest.permission.ACCESS_NETWORK_STATE)
                ) || BureauAPI.hasPermissions(context, listOf(Manifest.permission.READ_BASIC_PHONE_STATE))
            ) {
                if (telephonyManager.isDataRoamingEnabled) {
                    1
                } else {
                    0
                }
            } else {
                -1
            }
        }, -1)
    }
}
